import numpy
from numpy import *
from pylab import *

def square_equation():
    """Solve the equation x^2 - 4x + 3 = 0 (from 5.1) by fixed-point iteration.
    
    returns float as the output of a converged fixed-point iteration.

    Reference test:
    square_system() => 1.0
    """

    # Fill in

    return 0

def square_system():
    """Solve the equation system given in 5.1 by Newton's xmethod.
    
    returns array as the output of Newton's method.

    Reference test:
    square_system() => [ 1.51154714, -0.71522524]
    """

    # Fill in

    return 0

def timestep(f, t, u0, k, method, solver):
    """Take one time step for the initial value problem (IVP).

    f: function describing IVP u' = f(t, u)
    t: float for time value f(t, u)
    u0: float for initial value
    k: float for time-step value
    method: string naming the time-step method
    solver: string naming the algebraic solver
    ("fixedpoint" or "newton")
    
    Returns float u = u0 + k * f(t, u0)

    Reference test:
    def f(t, u): return t*t
    timestep(f, 0.1, 1.0, 0.1, "Euler") => 1.001
    """

    # Fill in

    return 0

def solve(f, I, u0, k, method, solver):
    """Solve the IVP f by time-stepping

    f: function describing IVP u' = f(t, u)
    I: float array for time interval
    u0: float for initial value
    k: float for time-step value
    method: string naming the time-step method
    solver: string naming the algebraic solver
    ("fixedpoint" or "newton")
    
    Returns array of arrays[tarray, uarray] for time values and u values

    Reference test:
    def f(t, u): return -u
    [tarray, uarray] = solve(f, [0.0, 3.8],
                             1.0, 1.9, "Midpoint", "fixedpoint")
    Expecting uarray[0, -1] = 0.00065 (6 digits tolerance)

    [tarray, uarray] = solve(f, [0.0, 40.0],
                             1.0, 4.0, "Midpoint", "newton")
    Expecting uarray[0, -1] = 1.69350e-05 (6 digits tolerance)"""

    # Fill in

    return 0

def fixedpoint(g, x0, data=[]):
    """Iterates the fixed-point iteration x = g(x) until
    convergence with the starting guess x0.

    x0: scalar/array of starting guess.
    g: function for fixed-point iteration x = g(x, iter, data)
      x: scalar/array
      iter: integer for iteration number
      data: list of extra data (parameters/objects) which can be used in the
            computation of g, a matrix or a decomposition of a matrix f.ex.
    

    """

    x = x0
    TOL = 1.0e-8
    res = 1.0
    iter = 0
    while(res > TOL):
        y = g(x, iter, data)
        r = y - x
        res = norm(r)
        x = y
        iter += 1

    print "residual:", res
    print "iterations: ", iter
    return x

def jacobian(f, u):
    """Computes the Jacobian matrix f'(u) using numerical
    differentiation.

    f: function f(u) to compute the Jacobian of.
    u: scalar/array to compute the Jacobian at.
    """

    y = f(u)
    if(size(u) > 1):
        J = zeros([size(y), size(u)])
    else:
        J = 0.0

    # Iterate over columns of J, computing one column at
    # a time.
    h = 1.0e-7
    for i in range(0, size(u)):
        hv = zeros([size(u)])
        hv[i] = h
        Jcol = (f(u + hv) - f(u)) / h

        if(size(u) > 1):
            J[:, i] = Jcol
        else:
            J = Jcol
        hv[i] = 0.0

    return J

def my_solve(A, b):
    """Solves the linear system Ax = b. Handles both
    A and b arrays, and A and b scalar.

    returns solution x of linear system.
    """

    if(size(b) > 1):
        x = numpy.linalg.solve(A, b)
    else:
        x = b / A
        x = x[0]

    return x

def newton(f, x0):
    """Solves the equation (system) f(x) = 0 by Newton's
    method with the starting guess x0.

    f: function describing equation f(x) = 0.
    x0: scalar/array of starting guess.
    """

    class LocalData:
        def __init__(self):
            self.f = f

    def g(x, iter, data):
        # Need to store f to compute the Jacobian matrix
        f = data.f
        J = jacobian(f, x)

        # Compute right hand side of Newton iteration
        # and solve the linear system.
        r = dot(J, x) - f(x)
        y = my_solve(J, r)

        return y

    data = LocalData()
    return fixedpoint(g, x0, data)

def fixedpoint_adapter(g):
    """ Converts a fixed-point function of the form:
    x = g(x) to the form f(x) = x - g(x) = 0.

    returns a function f(x) = 0
    """

    def f(x):
        return x - g(x, 0, [])

    return f

def jacobi(A, b):
    """Solves the linear system Ax = b with Jacobi's
    method.

    A: 2-dimensional numpy array for matrix A
    b: 1-dimensional numpy array for vector b

    returns array for solution x.

    Reference test:
    A = array([[70, 1, 5], [3, 80, 4], [6, 2, 90]], 'd')
    b = array([2, 4, 6])

    jacobi(A, b) => array([ 0.02333754,  0.04592032,  0.06409038])
    """

    # Fill in
    return 0

def steepest_descent(A, b):
    """Solves the linear system Ax = b with the Steepest
    Descent method.

    A: 2-dimensional array for matrix A
    b: 1-dimensional array for vector b

    returns array for solution x.
    """

    def g(x, iter, data):
        # Update direction d (equal to residual)
        r = b - dot(A, x)

        # Fill in

        return 0

    # Use b as initial guess for simplicity
    return fixedpoint(g, b)

def conjugate_gradient(A, b):
    """Solves the linear system Ax = b with the Conjugate
    Gradient method.

    A: 2-dimensional array for matrix A
    b: 1-dimensional array for vector b

    returns array for solution x.
    """

    class LocalData:
        # Store r, d, and alpha, since they are carried over
        # from the previous iteration.

        r = []
        d = []
        alpha = []

    def g(x, iter, data):
        r = data.r
        d = data.d
        alpha = data.alpha

        if(iter == 0):
            r = b - A*x
            d = r
        else:
            # Update direction d based on orth.
            # projection of r to Krylov space
            rold = r
            r = r - (A*d)*alpha
            beta = r.inner(r) / rold.inner(rold)
            d = r + d*beta

        # Compute step length
        alpha = r.inner(r) / d.inner(A*d)
        # Step/iterate in direction d with length alpha
        y = x + d*alpha

        data.r = r
        data.d = d
        data.alpha = alpha
        return y

    # Use b as initial guess for simplicity
    data = LocalData()
    return fixedpoint(g, b, data)
